/*
			Copyright (c) 1994 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/cut_tri.c#1 $
*/
#include <avs/util.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/fld.h>
/* #include <avs/dv_util.h> */

#define NEW_FLAG 1
#define OLD_FLAG 0

#define ERR_RETURN(A) ERRerror("cut_table",1,ERR_ORIG, A); return(0);
#define MAX_NAME_SIZE 1024


static FILE   *f_on, *f_nn, *f_nt, *f_min, *f_max, *f_fl, *f_con, *f_mind, *f_cind, *f_tind, *f_ct, *f_cn;
static int   m_ind, c_ind, t_ind, first=1, count = 0;
static int   on1=1, nn1=1, nt1=1, min1=1, fl1=1, con1=1, mind1=1, cind1=1, ct1=1;
static int   on0=0, nn0=0, nt0=0, min0=0, fl0=0, con0=0, mind0=0, cind0=0, ct0=0;

int FUNCcut_table (OMobj_id in, OMobj_id out);

static int print_con(int nnodes, int *tlist, int *flag, int nt,
                     int *ctype, int *cell_nnode);

cut_table(OMobj_id elem_id)
{
	OMobj_id in, out, level_id, param;
	double  level;

	in = OMfind_subobj(elem_id, OMstr_to_name("in"), OM_OBJ_RD);
	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
	if (FUNCcut_table(in, out)) {
		return(1);
	}
	else return(0);
}

int FUNCcut_table (OMobj_id in, OMobj_id out)
{
	int   nnodes, nspace,out_ncells, out_nnodes, data_type, size, conn_size, veclen, ncomp;
	int   *node_list, *out_nlist, *min_node_list, *max_node_list, stat;
	int   comp, *node_conn, ppnts, pconn[10], tlist[20], flag[20];
	int   ntet, *nlist, table_index, table_index1, n0,n1,n2,n3,n4,n5,n6,n7, nt, i, j, nn, max_tet;
	int   nfaces, *face_nnodes, *face_node_list;
	int   tnodes, ctype[20], cnnode[20];
	double null_value;
	char  units[MAX_NAME_SIZE];
	float  *coord, *out_coord, t[1000], iso_val=0.5, tol;
	float   *node_data, xyz[1000];
	float  min_ext[3], max_ext[3];
	char   *cell_table;
	OMobj_id   cell_set;


	null_value = 0.0;

	out_nnodes = 0;
	m_ind=0;
	c_ind=0;
	t_ind=0;
	if (!(f_on = fopen("f_on.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_nn = fopen("f_nn.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_nt = fopen("f_nt.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_min = fopen("f_min.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_max = fopen("f_max.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_fl = fopen("f_fl.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_con = fopen("f_con.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_cind = fopen("f_cind.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_mind = fopen("f_mind.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_tind = fopen("f_tind.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_ct = fopen("f_ct.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_cn = fopen("f_cn.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}

	if (FLDget_node_data_ncomp(in, &ncomp) != 1) {
		ERR_RETURN("Error getting ncomp");
	}
	if (ncomp == 0) {
		ERR_RETURN("bad number cut_table component index");
	}
	if (FLDget_node_data_veclen(in, 0, &veclen) !=1) {
		ERR_RETURN("Error getting veclen");
	}

	if (veclen != 1) {
		fprintf (stderr, "cut_table: first component is not scalar\n");
		return(0);
	}

	if (FLDget_nnodes(in, &nnodes) != 1) {
		ERR_RETURN("cannot get nnodes");
	}
	if (FLDget_nspace(in, &nspace) != 1) {
		ERR_RETURN("cannot get nnodes");
	}
	if (FLDget_node_data(in, 0, &data_type, (char **)&node_data, 
			      &size, OM_GET_ARRAY_RW) != 1) {
		ERR_RETURN("cannot get node data");
	}
	if (FLDget_coord(in, &coord, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("cannot get coordinates");
	}
	if (FLDget_coord_extent(in, min_ext, max_ext) != 1) {
		ERR_RETURN("cannot get extents");
	}
	stat = FLDget_mesh_info(in, &cell_table);
	if (!stat) {
		ERR_RETURN("cannot create cell table");
	}
	if (FLDget_cell_set(in, 0, &cell_set) != 1) {
		ERR_RETURN("cannot get cell set");
	}
	if (FLDget_node_connect(cell_set, &node_conn, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("cannot get connect");
	}

	fprintf (f_on, "static int CUTnold_nodes_tri[] = \n       {0,");
	fprintf (f_nn, "static int CUTnnew_nodes_tri[] = \n       {0,");
	fprintf (f_nt, "static int CUTncell_tri[] = \n       {0,");
	fprintf (f_min, "static unsigned char CUTmin_node_tri[] = \n       {");
	fprintf (f_max, "static unsigned char CUTmax_node_tri[] = \n       {");
	fprintf (f_mind, "static int CUTm_ind_tri[] = \n       {0,");
	fprintf (f_cind, "static int CUTc_ind_tri[] = \n       {0,");
	fprintf (f_tind, "static int CUTt_ind_tri[] = \n       {0,");
	fprintf (f_con, "static unsigned char CUTconnect_list_tri[] = \n       {");
	fprintf (f_fl, "static unsigned char CUTnew_flag_list_tri[] = \n       {");
	fprintf (f_ct, "static unsigned char CUTcell_type_tri[] = \n       {");
	fprintf (f_cn, "static unsigned char CUTcell_nnode_tri[] = \n       {");

	max_tet=0;
	for (n2=0; n2<2; n2++) {
		node_data[node_conn[2]] = (float)n2;
	for (n1=0; n1<2; n1++) {
		node_data[node_conn[1]] = (float)n1;
	for (n0=0; n0<2; n0++) {
		node_data[node_conn[0]] = (float)n0;

		ppnts = 0;

		if (node_data[node_conn[0]] >= iso_val) {
			memcpy(xyz+3*ppnts, coord+3*node_conn[0], 3*sizeof(float));
			pconn[ppnts++] = 0;
		}
		if (node_data[node_conn[1]] >= iso_val) {
			memcpy(xyz+3*ppnts, coord+3*node_conn[1], 3*sizeof(float));
			pconn[ppnts++] = 1;
		}
		if (node_data[node_conn[2]] >= iso_val) {
			memcpy(xyz+3*ppnts, coord+3*node_conn[2], 3*sizeof(float));
			pconn[ppnts++] = 2;
		}
		table_index = ((node_data[node_conn[2]] >= iso_val) << 2) |
			 ((node_data[node_conn[1]] >= iso_val) << 1) |
			 (node_data[node_conn[0]] >= iso_val);

		if ((table_index <= 0) || (table_index >= 7))
			continue;

		UTILisoline(cell_table, node_data, data_type, 
			    0, &null_value, (double)iso_val,
			    &out_nnodes, &min_node_list, &max_node_list,
			    &out_ncells, &out_nlist, &conn_size);

		UTILget_param(out_nnodes, data_type, node_data,
			      iso_val, min_node_list, max_node_list, t); 
		UTILeval_param(out_nnodes, 3, OM_TYPE_FLOAT, coord, 0, &null_value,
			       min_node_list, max_node_list, t, xyz+3*ppnts);

		UTILdelaun_tess_2D (ppnts+out_nnodes, xyz, nspace, min_ext, max_ext, 0, 
				    &ntet, &nlist);
		if (on1) {
			fprintf (f_on,"%d", ppnts);
			fprintf (f_nn,"%d", out_nnodes);
			fprintf (f_mind, "%d", m_ind);
			fprintf (f_cind, "%d", c_ind);
			fprintf (f_tind, "%d", t_ind);
			on1 = 0;
		}
		else {
			fprintf (f_on,",%d", ppnts);
			fprintf (f_nn,",%d", out_nnodes);
			fprintf (f_mind, ",%d", m_ind);
			fprintf (f_cind, ",%d", c_ind);
			fprintf (f_tind, ",%d", t_ind);
		}
		on0++;
		if (on0 >= 32) {
			fprintf (f_on, ",\n        ");
			fprintf (f_nn, ",\n        ");
			fprintf (f_mind, ",\n        ");
			fprintf (f_cind, ",\n        ");
			fprintf (f_tind, ",\n        ");
			on0 = 0;
			on1 = 1;
		}

		for (i=0; i<out_nnodes; i++) {
			if (min1) {
				fprintf (f_min, "%d",min_node_list[i]);
				fprintf (f_max, "%d",max_node_list[i]);
				m_ind++;
				min1=0;
			}
			else {
				fprintf (f_min, ",%d",min_node_list[i]);
				fprintf (f_max, ",%d",max_node_list[i]);
				m_ind++;
			}
			min0++;
			if (min0 >= 24) {
				fprintf (f_min, ",\n        ");
				fprintf (f_max, ",\n        ");
				min0 = 0;
				min1 = 1;
			}
		}
		for (nt = 0,nn=0; nn<ntet; nn++) {
			if (nlist[3*nn]>=ppnts &&
			    nlist[3*nn+1]>=ppnts &&
			    nlist[3*nn+2]>=ppnts)
				continue;
			for (i=0; i<3; i++) {
				if (nlist[3*nn+i]>=ppnts) {
					flag[i]=NEW_FLAG;
					tlist[i]=nlist[3*nn+i]-ppnts;
				}
				else {
					flag[i]=OLD_FLAG;
					tlist[i]=pconn[nlist[3*nn+i]];
				}
			}
			for (i=0; i<3; i++) {
				if (first) {
					fprintf (f_fl, "%d",flag[i]);
					fprintf (f_con, "%d",tlist[i]);
					first = 0;
				}
				else {
					fprintf (f_fl, ",%d",flag[i]);
					fprintf (f_con, ",%d",tlist[i]);
					}
				count++;
			}
			if (count >= 24) {
				fprintf (f_fl, ",\n        ");
				fprintf (f_con, ",\n        ");
				count = 0;
				first = 1;
			}
			c_ind += 3;
			nt++;
			if (ct1) {
				fprintf (f_ct, "      %d", 4);
				fprintf (f_cn, "      %d", 3);
				ct1 = 0;
			}
			else {
				fprintf (f_ct, ",%d", 4);
				fprintf (f_cn, ",%d", 3);
			}
			ct0++;
			if (ct0 >= 32) {
				fprintf (f_ct, ",\n        ");
				fprintf (f_cn, ",\n        ");
				ct0 = 0;
				ct1 = 1;
			}
			t_ind ++;
		}
		if (nt1) {
			fprintf (f_nt, "%d", nt);
			nt1 = 0;
		}
		else {
			fprintf (f_nt, ",%d", nt);
		}
		nt0++;
		if (nt0==24) {
			fprintf (f_nt, ",\n        ");
			nt1=1;
			nt0=0;
		}
		if (nt>max_tet)
			max_tet = nt;
		/**
		  fprintf(stderr," index=%d, index1=%d, old_points=%d, new_points=%d, ntet=%d\n",
		  table_index, table_index1, ppnts, out_nnodes, nt);
		  **/
	}}}


	fprintf (f_on, "};\n\n");
	fprintf (f_nt, "};\n\n   static int CUTmax_index_tri = %d;\n\n", table_index);
	fprintf (f_nn, "};\n\n");
	fprintf (f_min, "};\n\n");
	fprintf (f_max, "};\n\n");
	fprintf (f_mind, "};\n\n");
	fprintf (f_cind, "};\n\n");
	fprintf (f_tind, "};\n\n");
	fprintf (f_con, "};\n\n");
	fprintf (f_fl, "};\n\n");
	fprintf (f_ct, "};\n\n");
	fprintf (f_cn, "};\n\n");

	fclose(f_on);
	fclose(f_nn);
	fclose(f_fl);
	fclose(f_nt);
	fclose(f_min);
	fclose(f_max);
	fclose(f_con);
	fclose(f_mind);
	fclose(f_cind);
	fclose(f_tind);
	fclose(f_ct);
	fclose(f_cn);
	system("cat f_on.cut f_nn.cut f_mind.cut f_min.cut f_max.cut f_tind.cut f_ct.cut f_cn.cut f_nt.cut f_cind.cut f_con.cut f_fl.cut >cut_tbl.h");
	system("rm -rf f_on.cut f_nn.cut f_mind.cut f_min.cut f_max.cut f_tind.cut f_ct.cut f_cn.cut f_nt.cut f_cind.cut f_con.cut f_fl.cut");
	return(1);
}

static int print_con(int nnodes, int *tlist, int *flag, int nt,
                     int *ctype, int *cell_nnode)
{
	int  i,j;

	for (i=0; i<nnodes; i++) {
		if (first) {
			fprintf (f_fl, "%d",flag[i]);
			fprintf (f_con, "%d",tlist[i]);
			first = 0;
		}
		else {
			fprintf (f_fl, ",%d",flag[i]);
			fprintf (f_con, ",%d",tlist[i]);
		}
		count++;
	}
	if (count >= 24) {
		fprintf (f_fl, ",\n        ");
		fprintf (f_con, ",\n        ");
		count = 0;
		first = 1;
	}
	for (i=0; i<nt; i++) {
		if (ct1) {
			fprintf (f_ct, "      %d", ctype[i]);
			fprintf (f_cn, "      %d", cell_nnode[i]);
			ct1 = 0;
		}
		else {
			fprintf (f_ct, ",%d", ctype[i]);
			fprintf (f_cn, ",%d", cell_nnode[i]);
		}
		ct0 += nt;
		if (ct0 >= 32) {
			fprintf (f_ct, ",\n        ");
			fprintf (f_cn, ",\n        ");
			ct0 = 0;
			ct1 = 1;
		}
	}
	return(1);
}
